#!/usr/bin/env python3

import numpy as np

import os
import sys
import importlib
import textwrap
import numpy as np
import math as m
import multiprocessing as mp
import ast
import pprint
import importlib.util
import time
import math
import numbers


max_points      = 0
obtained_points = 0
default_permitted_modules = ["matplotlib",
                             "matplotlib.pyplot",
                             "matplotlib.widgets",
                             "functools",
                             "mpl_toolkits.mplot3d"]
timeout = 60
numeric_accuracy = 1e-6


def info(string):
    print(string)


def format_beautifully(*blocks):
    # step #1: convert short atomics to wrap statements
    preprocessed_blocks = []
    for (kind, object) in blocks:
        if   kind == 'wrap':
            preprocessed_blocks += [(kind, str(object))]
        elif (kind == 'atomic') or (kind == 'args'):
            string = pprint.pformat(object, width=79)
            if kind == 'args':
                string = "(" + string[1:-1] + ")"
            if '\n' in string:
                preprocessed_blocks += [('atomic', string)]
            else:
                preprocessed_blocks += [('wrap', string)]

    # step #2: merge consecutive wrap statements
    merged_blocks = []
    wrap_strings = []
    for (kind, string) in preprocessed_blocks:
        if   kind == 'wrap':
            wrap_strings += [string]
        elif kind == 'atomic':
            merged_blocks += [('wrap', "".join(wrap_strings))]
            wrap_strings = []
            merged_blocks += [(kind, string)]
    merged_blocks += [('wrap', "".join(wrap_strings))]

    # step #3: combine the text blocks
    #          in particular merge atomics followed by wrap text
    lines = []
    for (kind, string) in merged_blocks:
        if kind == 'wrap':
            if not lines:
                lines += textwrap.fill(string, width=79).splitlines()
            else:
                indent = len(lines[-1]) if lines else 0
                new_lines = textwrap.fill("#" * indent + string, width=79).splitlines()
                lines[-1] = new_lines[0].replace("#" * indent, lines[-1])
                lines += new_lines[1:]
        elif kind == 'atomic':
            lines += string.splitlines()

    return "\n".join(lines)


def type_equal(a,b):
    # surprise #1: numpy.int64 and the like are not of type integer!
    # surprise #2: numpy.float64 and Python float are disjoint!
    def denumpify(x):
        return np.asscalar(x) if isinstance(x, np.generic) else x
    return type(denumpify(a)) == type(denumpify(b))


def equal(a, b):
    if not type_equal(a,b):
        return False
    if isinstance(a, float):
        return abs(a - b) < numeric_accuracy
    if isinstance(a, np.poly1d):
        return equal(a.c,b.c)
    if isinstance(a, np.ndarray):
        if a.shape != b.shape: return False
        if a.size == 0: return True
        return np.amax(abs(a.astype(np.float64) - b.astype(np.float64))) < numeric_accuracy
    if isinstance(a, tuple) or isinstance(a, list):
        if len(a) != len(b): return False
        for (a, b) in zip(a,b):
            if not equal(a,b): return False
        return True
    return a == b


class CheckFailure(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return format_beautifully(('wrap', self.message))


class Timeout(CheckFailure):
    def __init__(self, call):
        self.call = call

    def __str__(self):
        (fname, *args) = self.call
        blocks = [('wrap', "Der Aufruf {}".format(fname)),
                  ('args', args),
                  ('wrap', " terminiert nicht, oder ist SEHR ineffizient.")]
        return format_beautifully(*blocks)


class WrongResult(CheckFailure):
    def __init__(self, call, expected_result, obtained_result):
        self.call = call
        self.expected_result = expected_result
        self.obtained_result = obtained_result

    def __str__(self):
        (fname, *args) = self.call
        blocks = [('wrap', "Der Aufruf {}".format(fname)),
                  ('args', args)]
        if isinstance(self.obtained_result, Exception):
            blocks += ([('wrap', " liefert den Fehler: "),
                        ('atomic', self.obtained_result)])
        elif isinstance(self.obtained_result, type(None)):
            blocks += ([('wrap', " hat keinen RÃ¼ckgabewert, sollte aber "),
                        ('atomic', self.expected_result),
                        ('wrap', " zurÃ¼ckgeben.")])
        elif not type_equal(self.obtained_result, self.expected_result):
            blocks += ([('wrap', " liefert ein Ergebnis vom Typ "),
                        ('wrap', "'" + type(self.obtained_result).__name__ + "'"),
                        ('wrap', ", erwartet wird aber ein Ergebnis vom Typ "),
                        ('wrap', "'" + type(self.expected_result).__name__ + "'.")])
        else:
            blocks += ([('wrap', " liefert das Ergebnis "),
                        ('atomic', self.obtained_result),
                        ('wrap', ", richtig wÃ¤re aber "),
                        ('atomic', self.expected_result),
                        ('wrap', ".")])
        return format_beautifully(*blocks)


class AstInspector(ast.NodeVisitor):
    def __init__(self, file="<unknown>", imports=[]):
        self.file = file
        self.permitted_modules = set(default_permitted_modules + imports)
        self.errors = []


    def check_module(self, module):
        if module in self.permitted_modules: return
        self.errors.append(CheckFailure("Das Modul {} darf fÃ¼r diese "
                                       "Aufgabe nicht verwendet werden."
                                       .format(module)))

    def visit_Import(self, node):
        for m in [n.name for n in node.names]:
            self.check_module(m)


    def visit_ImportFrom(self, node):
        self.check_module(node.module)


def subprocess_eval(queue, proc_id, expr, module):
    try:
        m = importlib.import_module(module)
        (f, *args) = map(lambda x: eval(x, vars(m)) if isinstance(x, str) else x, expr)
        result = f(*args)
    except BaseException as e:
        result = e
    queue.put((proc_id, result))


def check_calls(forms, module_name):
    calls, desired_results = zip(*forms) if forms else ([], [])
    queue = mp.Queue()
    processes = []
    process_ids = set(range(len(calls)))
    for pid in process_ids:
        p = mp.Process(target=subprocess_eval,
                       args=(queue, pid, calls[pid], module_name))
        p.start()
        processes.append(p)

    queue_contents = []
    try:
        for _ in process_ids:
            queue_contents.append(queue.get(True, timeout))
    except:
        pass
    finally:
        for p in processes: p.terminate()

    for pid in process_ids:
        match = [result for (id, result) in queue_contents if id == pid]
        if not match:
            yield Timeout(calls[pid])
        elif not equal(desired_results[pid], match[0]):
            yield WrongResult(calls[pid], desired_results[pid], match[0])


def check_module(module_name, imports=[], calls=[], nuke=[]):
    try:
        spec = importlib.util.find_spec(module_name)
        if not spec:
            yield CheckFailure("Die Datei {}.py konnte nicht gefunden werden. "
                              "Sie sollte im selben Ordner liegen wie "
                              "dieses Programm.".format(module_name))
            return

        with open(spec.origin, 'r') as f:
            source = f.read()
        st = ast.parse(source, spec.origin)
        inspector = AstInspector(spec.origin, imports)
        inspector.visit(st)
        errors = inspector.errors or check_calls(calls, module_name)
        for error in errors: yield error
    except OSError:
        yield CheckFailure("Die Datei {} konnte nicht geÃ¶ffnet werden."
                          .format(spec.origin))
    except SyntaxError as e:
        yield CheckFailure("Die Datei {} enthÃ¤lt einen Syntaxfehler "
                          "in Zeile {:d}."
                          .format(spec.origin, e.lineno))
    except CheckFailure as e:
        yield e
    except Exception as e:
        yield CheckFailure("Beim Laden des Moduls {} ist ein Fehler "
                          "vom Typ '{}' aufgetreten."
                          .format(module_name, str(e)))


def check(description, points, module, **kwargs):
    global max_points
    global obtained_points

    pstr = ("1 Punkt" if points == 1 else "{:d} Punkte".format(points))
    desc = description + " (" + pstr + ")"
    max_points += points
    info("=" * (len(desc) + 2))
    info(" " + desc)
    info("=" * (len(desc) + 2))

    try:
        stdout, stderr = sys.stdout, sys.stderr
        try:
            with open(os.devnull, 'w') as f:
                sys.stdout, sys.stderr = f, f
                errors = list(check_module(module, **kwargs))
        finally:
            sys.stdout, sys.stderr = stdout, stderr
    except BaseException as e:
        info("Eine unbehandelte Ausnahme ist aufgetreten: '{}'"
             .format(str(e)))
    else:
        if errors:
            for e in errors: info(str(e))
            info("Diese Teilaufgabe enthÃ¤lt Fehler!")
        elif not kwargs["calls"]:
            max_points -= points
            info("Diese Teilaufgabe hat keine Ã¶ffentlichen Testcases.")
        else:
            obtained_points += points
            info("Super! Alles scheint zu funktionieren.")
    info("")


def import_modules(modules):
    for module in modules:
        importlib.import_module(module)


def compute_process_spawn_time():
    start = time.time()
    p = mp.Process(target=import_modules, args=(default_permitted_modules,))
    p.start()
    p.join()
    end = time.time()
    return math.ceil(end - start)


def report():
    print("Insgesamt: {:d} von {:d} Punkten.".format(obtained_points, max_points))


if __name__ == "__main__":
    mp.freeze_support()
    timeout = 2 * compute_process_spawn_time() + 9


    ###############################
    ## Nun die eigentlichen Checks
    ###############################

    check("Aufgabe 1a: De Casteljau Schritt", 1, "bezier",
          imports = ["numpy"],
          calls = [
              (("de_casteljau_step", np.array([[0.,0.], [1.,0.]]), 0.5), np.array([[0.5, 0]])),
              (("de_casteljau_step", np.array([[0.,0.], [1.,0.]]), 0.2), np.array([[0.2, 0]])),
              (("de_casteljau_step", np.array([[0.,0.], [1.,1.]]), 0.5), np.array([[0.5, 0.5]])),
              (("de_casteljau_step", np.array([[0.,0.], [1.,1.]]), 0.2), np.array([[0.2, 0.2]])),
              (("de_casteljau_step", np.array([[0.,0.], [1.,1.], [2.,2.]]), 0.5), np.array([[0.5, 0.5], [1.5, 1.5]])),
              (("de_casteljau_step", np.array([[0.,0.], [1.,1.], [2.,0.]]), 0.5), np.array([[0.5, 0.5], [1.5, 0.5]])),
              (("de_casteljau_step", np.array([[0.,0.], [1.,1.], [2.,2.], [3.,3.]]), 0.2),
               np.array([[0.2, 0.2], [1.2, 1.2], [2.2,2.2]])),
              (("de_casteljau_step", np.array([[3., 2.], [2., 5.], [7., 6.], [8., 1.]]), 0.3),
                np.array([[2.7, 2.9], [3.5, 5.3], [7.3, 4.5]])),
          ])

    check("Aufgabe 1b: De Casteljau Algorithmus", 1, "bezier",
          imports = ["numpy"],
          calls = [
              (("de_casteljau", np.array([[0,0], [1,0]]), 0.5), np.array([0.5, 0])),
              (("de_casteljau", np.array([[0,0], [1,0]]), 0.2), np.array([0.2, 0])),
              (("de_casteljau", np.array([[0.,0.], [1.,1.]]), 0.2), np.array([0.2, 0.2])),
              (("de_casteljau", np.array([[0.,0.], [1.,1.], [2.,2.]]), 0.5), np.array([1.,1.])),
              (("de_casteljau", np.array([[0.,0.], [1.,1.], [2.,0.]]), 0.5), np.array([1.,0.5])),
              (("de_casteljau", np.array([[0.,0.], [1.,1.], [2.,2.], [3.,3.]]), 0.5),
               np.array([1.5,1.5])),
              (("de_casteljau", np.array([[0.,0.], [1.,1.], [2.,2.], [3.,3.]]), 0.2),
               np.array([0.6,0.6])),
              (("de_casteljau", np.array([[0.,0.], [1.,1.], [2.,2.], [3.,3.], [4.,4.]]), 0.2),
               np.array([0.8,0.8])),
              (("de_casteljau", np.array([[0.,0.], [1.,1.], [2.,2.], [1.,1.], [0.,0.]]), 0.5),
               np.array([1.25,1.25])),
              (("de_casteljau", np.array([[3., 2.], [2., 5.], [7., 6.], [8., 1.]]), 0.3),
                np.array([3.45, 4.052])),
          ])

    check("Aufgabe 1c: BÃ©zierkurven approximieren", 1, "bezier",
          imports = ["numpy"],
          calls = [
              (("bezier1", np.array([[0.,0.], [1.,0.]]), 2), np.array([[0.0, 0.0], [1.0, 0.0]])),
              (("bezier1", np.array([[0.,0.], [1.,1.]]), 2), np.array([[0.0, 0.0], [1.0, 1.0]])),
              (("bezier1", np.array([[0.,0.], [1.,0.]]), 3), np.array([[0.0, 0.0], [0.5,0.0], [1.0, 0.0]])),
              (("bezier1", np.array([[0.,0.], [1.,1.]]), 3), np.array([[0.0, 0.0], [0.5,0.5], [1.0, 1.0]])),
              (("bezier1", np.array([[0.,0.], [1.,1.], [2., 0.]]), 2), np.array([[0., 0.], [2., 0.]])),
              (("bezier1", np.array([[0.,0.], [1.,1.], [2., 0.]]), 3), np.array([[0., 0.], [1., 0.5], [2., 0.]])),
              (("bezier1", np.array([[0.,0.], [1.,1.], [2., 0.]]), 5),
               np.array([[0., 0.], [0.5, 0.375], [1., 0.5], [1.5, 0.375], [2., 0.]])),
              (("bezier1", np.array([[3., 2.], [2., 5.], [7., 6.], [8., 1.]]), 3),
                np.array([[3., 2.], [4.75, 4.5], [8., 1.]])),
              (("bezier1", np.array([[3., 2.], [3.04, 3.528], [4., 6.], [8., 1.]]), 10),
               np.array([[ 3.        ,  2.        ],
                         [ 3.0503155 ,  2.53275171],
                         [ 3.18622771,  3.06616187],
                         [ 3.42518519,  3.53096296],
                         [ 3.78463649,  3.85788752],
                         [ 4.28203018,  3.97766804],
                         [ 4.93481481,  3.82103704],
                         [ 5.76043896,  3.31872702],
                         [ 6.77635117,  2.40147051],
                         [ 8.        ,  1.        ]])),
          ])

    check("Aufgabe 1d: Kontrollpunkte hinzufÃ¼gen", 2, "bezier",
          imports = ["numpy"],
          calls = [
              (("add_control_point", np.array([[0.,0.], [1.,0.]])), np.array([[0., 0.], [0.5,0.], [1.,0.]])),
              (("add_control_point", np.array([[0.,0.], [0.5, 0.], [1.,0.]])), np.array([[0., 0.], [1/3,0.], [2/3, 0.], [1.,0.]])),
              (("add_control_point", np.array([[0.,0.], [1/3, 0.], [2/3, 0.], [1.,0.]])),
               np.array([[0., 0.], [0.25, 0.], [0.5, 0.], [0.75, 0.], [1., 0.]])),
              (("add_control_point", np.array([[0.,0.], [1/3, 1.], [2/3, 2.], [1.,3.]])),
               np.array([[0., 0.], [0.25, 0.75], [0.5, 1.5], [0.75, 2.25], [1., 3.]])),
              (("add_control_point", np.array([[3., 2.], [2., 5.], [7., 6.], [8., 1.]])),
                np.array([[3., 2.], [2.25, 4.25], [4.5, 5.5], [7.25, 4.75], [8., 1.]])),
          ])

    check("Aufgabe 1e: BÃ©zierkurven unterteilen", 2, "bezier",
          imports = ["numpy"],
          calls = [
              (("split_curve", np.array([[0.,0.], [1.,0.]])),
               (np.array([[0., 0.], [0.5, 0.]]), np.array([[0.5, 0.], [1., 0.]]))),
              (("split_curve", np.array([[0.,0.], [1.,1.]])),
               (np.array([[0., 0.], [0.5, 0.5]]), np.array([[0.5, 0.5], [1., 1.]]))),
              (("split_curve", np.array([[0.,0.], [1.,1.], [2., 0.]])),
               (np.array([[0., 0.], [0.5, 0.5], [1., 0.5]]), np.array([[1., 0.5], [1.5, 0.5], [2., 0.]]))),
              (("split_curve", np.array([[3., 2.], [2., 5.], [7., 6.], [8., 1.]])),
               (np.array([[3.,2.], [2.5, 3.5], [3.5, 4.5], [4.75, 4.5]]),
                np.array([[4.75, 4.5], [6., 4.5], [7.5, 3.5], [8., 1.]]))),
          ])

    check("Aufgabe 1f: BÃ©zierkurven rekursiv approximieren", 2, "bezier",
          imports = ["numpy"],
          calls = [
              (("bezier2", np.array([[0.,0.], [1.,0.]]), 0), np.array([[0.,0.], [1.,0.]])),
              (("bezier2", np.array([[0.,0.], [1.,0.]]), 1), np.array([[0.,0.], [0.5, 0.], [1.,0.]])),
              (("bezier2", np.array([[0.,0.], [1.,0.]]), 2),
               np.array([[ 0.  ,  0.  ],
                         [ 0.25,  0.  ],
                         [ 0.5 ,  0.  ],
                         [ 0.75,  0.  ],
                         [ 1.  ,  0.  ]])),
              (("bezier2", np.array([[0.,0.], [1.,0.]]), 3),
               np.array([[ 0.   ,  0.   ],
                         [ 0.125,  0.   ],
                         [ 0.25 ,  0.   ],
                         [ 0.375,  0.   ],
                         [ 0.5  ,  0.   ],
                         [ 0.625,  0.   ],
                         [ 0.75 ,  0.   ],
                         [ 0.875,  0.   ],
                         [ 1.   ,  0.   ]])),
              (("bezier2", np.array([[3., 2.], [2., 5.], [6., 6.], [8., 1.]]), 1),
               np.array([[ 3.   ,  2.   ],
                         [ 2.5  ,  3.5  ],
                         [ 3.25 ,  4.5  ],
                         [ 4.375,  4.5  ],
                         [ 5.5  ,  4.5  ],
                         [ 7.   ,  3.5  ],
                         [ 8.   ,  1.   ]])),
              (("bezier2", np.array([[3., 2.], [2., 5.], [6., 6.], [8., 1.]]), 2),
               np.array([[ 3.      ,  2.      ],
                         [ 2.75    ,  2.75    ],
                         [ 2.8125  ,  3.375   ],
                         [ 3.078125,  3.8125  ],
                         [ 3.34375 ,  4.25    ],
                         [ 3.8125  ,  4.5     ],
                         [ 4.375   ,  4.5     ],
                         [ 4.9375  ,  4.5     ],
                         [ 5.59375 ,  4.25    ],
                         [ 6.234375,  3.6875  ],
                         [ 6.875   ,  3.125   ],
                         [ 7.5     ,  2.25    ],
                         [ 8.      ,  1.      ]])
               ),
          ])


    report()